/*
  tidy2.c - HTML TidyLib command line driver

  Copyright (c) 1998-2007 World Wide Web Consortium
  (Massachusetts Institute of Technology, European Research 
  Consortium for Informatics and Mathematics, Keio University).
  All Rights Reserved.

  This version is with an 'enhanced' command processor, where
  (a) Supports @input.fil, with line delimited options,
      comments start with ';' (new)
  (b) Only support ONE command line set at a time (new behaviour)
  (c) Will ABORT (exit) on any command error (new behaviour)
  (d) All command line checks are from expanded tables (new)
  (e) Exit with -
      0 = no errors or warnings in parse, or any HELP command
      1 = some warnings
      2 = some errors, and maybe warnings
      3 = command parse error (new behaviour)
  (f) If ANY windows '/', give HELP (new behaviour)

   There is one anomaly. Old tidy supported
   -indent, or -i, to set indent to auto only
   *AND*
   --indent <columns>
   If TIDY_NEW_INDENT is undefined below, then this tidy emulates that behaviour.

   However, if TIDY_NEW_INDENT is defined, then NO LONGER will -indent set auto,
   and -i can NOT be used in the set of single character options.
   That is ALL FORMS of 'indent' require a 2nd tri-state argument, of auto, yes, or no.

  CVS (pseudo) Info :

    $Author: geoffmc $ 
    $Date: 2008/01/09 12:00:00 $ 
    $Revision: 1.48 $ 

*/

#pragma warning( disable:4090 ) /* different 'const' qualifier */
#include "tidy.h"
#include <errno.h>
#include <direct.h>  /* for _getcwd(buf,len) */

#define  TIDY_NEW_INDENT
#if   defined(UNICODE) || defined(_UNICODE)
#pragma message( "UNICODE Version ..." )
#else
#pragma message( "NON-UNICODE Version ..." )
#endif

static void pgm_exit( TidyDoc tdoc, int ex );
#define  CHKMEM(a)   if( !a ) { fprintf( errout, "ERROR: MEMORY FAILED!\n" ); pgm_exit( 0, 3 ); }

typedef Bool (*CMDSERVB)( TidyDoc, TidyOptionId, Bool  );
typedef Bool (*CMDSERV2)( TidyDoc, tmbstr );

/* static data variables */
#ifdef   TIDY_NEW_INDENT
static tmbstr single_letters = "bcemnqu"; /* NO 'i' here */
#else /* !#ifdef   TIDY_NEW_INDENT */
static tmbstr single_letters = "bceimnqu";
#endif /* #ifdef   TIDY_NEW_INDENT y/n */

static tmbstr prog = "dummy";
static FILE* errout = NULL;  /* set to stderr */
static ctmbstr cfgfil = NULL;
static ctmbstr errfil = NULL;
static ctmbstr htmlfil = NULL;
/* static FILE* txtout = NULL; */  /* set to stdout */
#ifdef   WIN32
#define  TIDY_MAX_PATH  264
static ctmbstr homedrive = NULL;
static ctmbstr homepath = NULL;
static tmbchar tidyrc[TIDY_MAX_PATH];
#endif   // WIN32

typedef enum
{
  CmdOptFileManip,
  CmdOptCatFIRST = CmdOptFileManip,
  CmdOptProcDir,
  CmdOptCharEnc,
  CmdOptMisc,
  CmdOptCatLAST
} CmdOptCategory;

static const struct {
    ctmbstr mnemonic;
    ctmbstr name;
} cmdopt_catname[] = {
    { "file-manip", "File manipulation" },
    { "process-directives", "Processing directives" },
    { "char-encoding", "Character encodings" },
    { "misc", "Miscellaneous" }
};

typedef struct {
   CMDSERVB  servb;     /* server */
   TidyOptionId optid;  /* option ID */
   CMDSERV2  serv2;     /* server with 2nd argument */
   CMDSERV2  serv3;     /* server with 1st as argument */
   uint     flag;       /* for output checking */
}CMDTABLE1, * PCMDTABLE1;

typedef struct {
    ctmbstr name1;      /**< Name */
    ctmbstr desc;       /**< Description */
    ctmbstr eqconfig;   /**< Equivalent configuration option */
    CmdOptCategory cat; /**< Category */
    ctmbstr name2;      /**< Name */
    ctmbstr name3;      /**< Name */
    CMDTABLE1 cmdtable;  /**< Command table entries */
} CmdOptDesc;

Bool tidy_SetIndent( TidyDoc tdoc, TidyOptionId id, Bool b );
Bool tidy_NewIndent( TidyDoc tdoc, tmbstr arg2 );
Bool tidy_SetBool( TidyDoc tdoc, TidyOptionId id, Bool b );
Bool tidy_UnsetBool( TidyDoc tdoc, TidyOptionId id, Bool b );
Bool tidy_Help( TidyDoc tdoc, TidyOptionId id, Bool b );
Bool tidy_Xml_Help( TidyDoc tdoc, TidyOptionId id, Bool b );
Bool tidy_Conf_Help( TidyDoc tdoc, TidyOptionId id, Bool b );
Bool tidy_Xml_Conf( TidyDoc tdoc, TidyOptionId id, Bool b );
Bool tidy_Shw_Conf( TidyDoc tdoc, TidyOptionId id, Bool b );
void tidy_Check_Arg2( TidyDoc tdoc, tmbstr arg2 );
void  tidy_Fix_Error_File( TidyDoc tdoc );
Bool tidy_Conf_File( TidyDoc tdoc, tmbstr arg2 );
Bool tidy_Set_Lang( TidyDoc tdoc, tmbstr arg2 );
Bool tidy_Set_Out( TidyDoc tdoc, tmbstr arg2 );
Bool tidy_Set_Err( TidyDoc tdoc, tmbstr arg2 );
Bool tidy_Set_Wrap( TidyDoc tdoc, tmbstr arg2 );
Bool tidy_Show_Vers( TidyDoc tdoc, tmbstr arg2 );
Bool tidy_Set_Access( TidyDoc tdoc, tmbstr arg2 );
Bool tidy_Set_Encoding( TidyDoc tdoc, tmbstr arg );
CmdOptDesc cmdopt_defs[];
CmdOptDesc * Tidy_Get_Cmd_Opt_Ptr( tmbstr arg );

/* #ifndef  NDEBUG */
static tmbstr tidy_strdupe( TidyDoc tdoc, tmbstr stg );
typedef struct _CMDLIST {
   void * link;
   tmbchar   cmd[1];
}CMDLIST, * PCMDLIST;

PCMDLIST pcmdlist = NULL;
PCMDLIST get_end_ptr( void )
{
   PCMDLIST pcl;
   PCMDLIST pcllast;
   pcllast = NULL;
   for ( pcl = pcmdlist; pcl; pcl = pcl->link )
   {
      pcllast = pcl;
   }
   return pcllast;
}

int get_pcl_count( void )
{
   PCMDLIST pcl;
   int   cnt = 0;
   for ( pcl = pcmdlist; pcl; pcl = pcl->link )
   {
      cnt++;
   }
   return cnt;
}

PCMDLIST get_pcl_item( int i )
{
   PCMDLIST pcl;
   int   cnt = 0;
   for ( pcl = pcmdlist; pcl; pcl = pcl->link )
   {
      if ( i == cnt )
         return pcl;
      cnt++;
   }
   return pcl;
}

static void save_commands( TidyDoc tdoc, int argc, char * * argv )
{
   int i;
   for ( i = 1; i < argc; i++ )
   {
      size_t len = strlen( argv[i] );
      PCMDLIST pcl = (PCMDLIST)malloc( sizeof(CMDLIST) + len );
      CHKMEM(pcl);
      strcpy( pcl->cmd, argv[i] );
      pcl->link = NULL;
      if ( pcmdlist == NULL )
      {
         pcmdlist = pcl;
      }
      else
      {
         PCMDLIST pcll = get_end_ptr();
         pcll->link = pcl;
      }
   }
}

static void show_commands( TidyDoc tdoc )
{
   PCMDLIST pcl;
   int   cnt = 0;
   for ( pcl = pcmdlist; pcl; pcl = pcl->link )
   {
      cnt++;
   }
   if ( cnt )
   {
      fprintf( errout, "Following %d commands ...\n", cnt );
      for ( pcl = pcmdlist; pcl; pcl = pcl->link )
      {
         fprintf( errout, "%s ", pcl->cmd );
      }
      fprintf( errout, "\n" );
   }
}

/* #endif /* #ifndef  NDEBUG */


static void pgm_exit( TidyDoc tdoc, int ex )
{
/* #ifndef  NDEBUG */
   if ( ex == 3 )
      show_commands( tdoc );
/* #endif /* #ifndef  NDEBUG */
   if ( tdoc )
      tidyRelease( tdoc );

   if ( ex == 3 ) {
      fprintf( errout, "ERROR: ABORTING APPLICATION! Error Level 3!\n" );
   } else {
#ifndef  NDEBUG
      printf( "Doing pgm_exit: Error Level %d ...\n", ex );
#endif   /* #ifndef  NDEBUG */
   }
   exit( ex );
}


static Bool samefile( ctmbstr filename1, ctmbstr filename2 )
{
#if FILENAMES_CASE_SENSITIVE
    return ( strcmp( filename1, filename2 ) == 0 );
#else
    return ( strcasecmp( filename1, filename2 ) == 0 );
#endif
}

static const char *cutToWhiteSpace(const char *s, uint offset, char *sbuf)
{
    if (!s)
    {
        sbuf[0] = '\0';
        return NULL;
    }
    else if (strlen(s) <= offset)
    {
        strcpy(sbuf,s);
        sbuf[offset] = '\0';
        return NULL;
    }
    else
    {
        uint j, l, n;
        j = offset;
        while(j && s[j] != ' ')
            --j;
        l = j;
        n = j+1;
        /* no white space */
        if (j==0)
        {
            l = offset;
            n = offset;
        }
        strncpy(sbuf,s,l);
        sbuf[l] = '\0';
        return s+n;
    }
}

static void print2Columns( const char* fmt, uint l1, uint l2,
                           const char *c1, const char *c2 )
{
    const char *pc1=c1, *pc2=c2;
    char *c1buf = (char *)malloc(l1+1);
    char *c2buf = (char *)malloc(l2+1);

    do
    {
        pc1 = cutToWhiteSpace(pc1, l1, c1buf);
        pc2 = cutToWhiteSpace(pc2, l2, c2buf);
        printf(fmt,
               c1buf[0]!='\0'?c1buf:"",
               c2buf[0]!='\0'?c2buf:"");
    } while (pc1 || pc2);
    free(c1buf);
    free(c2buf);
}

static void print3Columns( const char* fmt, uint l1, uint l2, uint l3,
                           const char *c1, const char *c2, const char *c3 )
{
    const char *pc1=c1, *pc2=c2, *pc3=c3;
    char *c1buf = (char *)malloc(l1+1);
    char *c2buf = (char *)malloc(l2+1);
    char *c3buf = (char *)malloc(l3+1);

    do
    {
        pc1 = cutToWhiteSpace(pc1, l1, c1buf);
        pc2 = cutToWhiteSpace(pc2, l2, c2buf);
        pc3 = cutToWhiteSpace(pc3, l3, c3buf);
        printf(fmt,
               c1buf[0]!='\0'?c1buf:"",
               c2buf[0]!='\0'?c2buf:"",
               c3buf[0]!='\0'?c3buf:"");
    } while (pc1 || pc2 || pc3);
    free(c1buf);
    free(c2buf);
    free(c3buf);
}

static const char helpfmt[] = " %-19.19s %-58.58s\n";
static const char helpul[]
        = "-----------------------------------------------------------------";
static const char fmt[] = "%-27.27s %-9.9s  %-40.40s\n";
static const char valfmt[] = "%-27.27s %-9.9s %-1.1s%-39.39s\n";
static const char ul[]
        = "=================================================================";


/* static const */
CmdOptDesc cmdopt_defs[] =  {
    { "-output <file>",
      "write output to the specified <file>",
      "output-file: <file>", CmdOptFileManip, "-o <file>", 0, /* add -output-file??? */
      { 0,              0,    tidy_Set_Out,   0, 0 } },

    { "-config <file>",
      "set configuration options from the specified <file>",
      NULL, CmdOptFileManip, 0, 0,
      { 0,              0,    tidy_Conf_File, 0, 0 } },

    { "-file <file>",
      "write errors and warnings to the specified <file>",
      "error-file: <file>", CmdOptFileManip, "-f <file>", 0,
      { 0,              0,    tidy_Set_Err,   0, 0 } },

    { "-modify",
      "modify the original input files",
      "write-back: yes", CmdOptFileManip, "-m", 0,   /* obsolete -change */
      { tidy_SetBool,   TidyWriteBack,     0, 0, 0 } }, /* added m */

    { "-indent",
      "indent element content",
      "indent: auto", CmdOptProcDir, "-i", 0,
#ifdef  TIDY_NEW_INDENT
      { 0, 0, tidy_NewIndent, 0, 0 } }, /* added i */
#else /* !#ifdef  TIDY_NEW_INDENT */
      { tidy_SetIndent, TidyIndentContent, 0, 0, 0 } }, /* added i */
#endif /* #ifdef  TIDY_NEW_INDENT y/n */

    { "-wrap <column>",
      "wrap text at the specified <column>"
      ". 0 is assumed if <column> is missing. "
      "When this option is omitted, the default of the configuration option "
      "\"wrap\" applies.",
      "wrap: <column>", CmdOptProcDir, "-w <column>", 0,
      { 0,              0,    tidy_Set_Wrap,  0, 0 } },

    { "-upper",
      "force tags to upper case",
      "uppercase-tags: yes", CmdOptProcDir, "-u", 0,
      { tidy_SetBool,   TidyUpperCaseTags, 0, 0, 0 } }, /* added u */

    { "-clean",
      "replace FONT, NOBR and CENTER tags by CSS",
      "clean: yes", CmdOptProcDir, "-c", 0,
      { tidy_SetBool,   TidyMakeClean,     0, 0, 0 } }, /* added c */

    { "-bare",
      "strip out smart quotes and em dashes, etc.",
      "bare: yes", CmdOptProcDir, "-b", 0,
      { tidy_SetBool,   TidyMakeBare,      0, 0, 0 } }, /* added b */

    { "-numeric",
      "output numeric rather than named entities",
      "numeric-entities: yes", CmdOptProcDir, "-n", 0,
      { tidy_SetBool,   TidyNumEntities,   0, 0, 0 } }, /* added n */

    { "-errors",
      "show only errors and warnings",
      "markup: no", CmdOptProcDir, "-e", 0,
      { tidy_UnsetBool, TidyShowMarkup,    0, 0, 0 } }, /* added e */

    { "-quiet",
      "suppress nonessential output",
      "quiet: yes", CmdOptProcDir, "-q", 0,
      { tidy_SetBool,   TidyQuiet,         0, 0, 0 } }, /* added q */

    { "-omit",
      "omit optional end tags",
      "hide-endtags: yes", CmdOptProcDir, 0, 0,
      { tidy_SetBool,   TidyHideEndTags,   0, 0, 0 } }, 

    { "-xml",
      "specify the input is well formed XML",
      "input-xml: yes", CmdOptProcDir, 0, 0,
      { tidy_SetBool,   TidyXmlTags,       0, 0, 0 } },

    { "-asxml",
      "convert HTML to well formed XHTML",
      "output-xhtml: yes", CmdOptProcDir, "-asxhtml", 0,
      { tidy_SetBool,   TidyXhtmlOut,      0, 0, 0 } },

    { "-ashtml",
      "force XHTML to well formed HTML",
      "output-html: yes", CmdOptProcDir, 0, 0,
      { tidy_SetBool,   TidyHtmlOut,       0, 0, 0 } }, 

#if SUPPORT_ACCESSIBILITY_CHECKS
    { "-access <level>",
      "do additional accessibility checks (<level> = 0, 1, 2, 3)"
      ". 0 is assumed if <level> is missing.",
      "accessibility-check: <level>", CmdOptProcDir, 0, 0,
      { 0,              0,    tidy_Set_Access,0, 0 } },
#endif

    { "-raw",
      "output values above 127 without conversion to entities",
      NULL, CmdOptCharEnc, 0, 0,
      { 0,              0, 0, tidy_Set_Encoding, 0 } },

    { "-ascii",
      "use ISO-8859-1 for input, US-ASCII for output",
      NULL, CmdOptCharEnc, 0, 0,
      { 0,              0, 0, tidy_Set_Encoding, 0 } },

    { "-latin0",
      "use ISO-8859-15 for input, US-ASCII for output",
      NULL, CmdOptCharEnc, 0, 0,
      { 0,              0, 0, tidy_Set_Encoding, 0 } },

    { "-latin1",
      "use ISO-8859-1 for both input and output",
      NULL, CmdOptCharEnc, 0, 0,
      { 0,              0, 0, tidy_Set_Encoding, 0 } },

#ifndef NO_NATIVE_ISO2022_SUPPORT
    { "-iso2022",
      "use ISO-2022 for both input and output",
      NULL, CmdOptCharEnc, 0, 0,
      { 0,              0, 0, tidy_Set_Encoding, 0 } },
#endif

    { "-utf8",
      "use UTF-8 for both input and output",
      NULL, CmdOptCharEnc, 0, 0,
      { 0,              0, 0, tidy_Set_Encoding, 0 } },

    { "-mac",
      "use MacRoman for input, US-ASCII for output",
      NULL, CmdOptCharEnc, 0, 0,
      { 0,              0, 0, tidy_Set_Encoding, 0 } },

    { "-win1252",
      "use Windows-1252 for input, US-ASCII for output",
      NULL, CmdOptCharEnc, 0, 0,
      { 0,              0, 0, tidy_Set_Encoding, 0 } },

    { "-ibm858",
      "use IBM-858 (CP850+Euro) for input, US-ASCII for output",
      NULL, CmdOptCharEnc, 0, 0,
      { 0,              0, 0, tidy_Set_Encoding, 0 } },

#if SUPPORT_UTF16_ENCODINGS
    { "-utf16le",
      "use UTF-16LE for both input and output",
      NULL, CmdOptCharEnc, 0, 0,
      { 0,              0, 0, tidy_Set_Encoding, 0 } },

    { "-utf16be",
      "use UTF-16BE for both input and output",
      NULL, CmdOptCharEnc, 0, 0,
      { 0,              0, 0, tidy_Set_Encoding, 0 } },

    { "-utf16",
      "use UTF-16 for both input and output",
      NULL, CmdOptCharEnc, 0, 0,
      { 0,              0, 0, tidy_Set_Encoding, 0 } },
#endif

#if SUPPORT_ASIAN_ENCODINGS /* #431953 - RJ */
    { "-big5",
      "use Big5 for both input and output",
      NULL, CmdOptCharEnc, 0, 0,
      { 0,              0, 0, tidy_Set_Encoding, 0 } },

    { "-shiftjis",
      "use Shift_JIS for both input and output",
      NULL, CmdOptCharEnc, 0, 0,
      { 0,              0, 0, tidy_Set_Encoding, 0 } },

    { "-language <lang>",
      "set the two-letter language code <lang> (for future use)",
      "language: <lang>", CmdOptCharEnc, 0, 0,  /* alternate -lang ??? */
      { 0,              0,    tidy_Set_Lang,  0, 0 } },
#endif

    { "-version",
      "show the version of Tidy",
      NULL, CmdOptMisc, "-v", 0,
      { 0,              0,    tidy_Show_Vers, 0, 0 } },

    { "-help",
      "list the command line options",
      NULL, CmdOptMisc, "-h", "-?",
      { tidy_Help,      0,                 0, 0, 0 } },

    { "-xml-help",
      "list the command line options in XML format",
      NULL, CmdOptMisc, 0, 0,
      { tidy_Xml_Help,  0,                 0, 0, 0 } },

    { "-help-config",
      "list all configuration options",
      NULL, CmdOptMisc, 0, 0,
      { tidy_Conf_Help, 0,                 0, 0, 0 } },

    { "-xml-config",
      "list all configuration options in XML format",
      NULL, CmdOptMisc, 0, 0,
      { tidy_Xml_Conf,  0,                 0, 0, 0 } },

    { "-show-config",
      "list the current configuration settings",
      NULL, CmdOptMisc, 0, 0,
      { tidy_Shw_Conf,  0,                 0, 0, 0 } },

    { NULL, NULL, NULL, CmdOptMisc, 0, 0,
      { 0,              0, 0, 0,                 0 } }
};

static tmbstr get_option_names( const CmdOptDesc* pos )
{
    tmbstr name;
    uint len = strlen(pos->name1);
    if (pos->name2)
        len += 2+strlen(pos->name2);
    if (pos->name3)
        len += 2+strlen(pos->name3);

    name = (tmbstr)malloc(len+1);
    strcpy(name, pos->name1);
    if (pos->name2)
    {
        strcat(name, ", ");
        strcat(name, pos->name2);
    }
    if (pos->name3)
    {
        strcat(name, ", ");
        strcat(name, pos->name3);
    }
    return name;
}

static tmbstr get_escaped_name( ctmbstr name )
{
    tmbstr escpName;
    char aux[2];
    uint len = 0;
    ctmbstr c;
    for(c=name; *c!='\0'; ++c)
        switch(*c)
        {
        case '<':
        case '>':
            len += 4;
            break;
        case '"':
            len += 6;
            break;
        default:
            len += 1;
            break;
        }

    escpName = (tmbstr)malloc(len+1);
    escpName[0] = '\0';

    aux[1] = '\0';
    for(c=name; *c!='\0'; ++c)
        switch(*c)
        {
        case '<':
            strcat(escpName, "&lt;");
            break;
        case '>':
            strcat(escpName, "&gt;");
            break;
        case '"':
            strcat(escpName, "&quot;");
            break;
        default:
            aux[0] = *c;
            strcat(escpName, aux);
            break;
        }

    return escpName;
}

static void print_help_option( void )
{
    CmdOptCategory cat = CmdOptCatFIRST;
    CmdOptDesc* pos = cmdopt_defs;

    for( cat=CmdOptCatFIRST; cat!=CmdOptCatLAST; ++cat)
    {
        size_t len =  strlen(cmdopt_catname[cat].name);
        printf("%s\n", cmdopt_catname[cat].name );
        printf("%*.*s\n", (int)len, (int)len, helpul );
        for( pos=cmdopt_defs; pos->name1; ++pos)
        {
            tmbstr name;
            if (pos->cat != cat)
                continue;
            name = get_option_names( pos );
            print2Columns( helpfmt, 19, 58, name, pos->desc );
            free(name);
        }
        printf("\n");
    }
}

static void print_xml_help_option_element( ctmbstr element, ctmbstr name )
{
    tmbstr escpName;
    if (!name)
        return;
    printf("  <%s>%s</%s>\n", element, escpName = get_escaped_name(name),
           element);
    free(escpName);
}

static void print_xml_help_option( void )
{
    CmdOptDesc* pos = cmdopt_defs;

    for( pos=cmdopt_defs; pos->name1; ++pos)
    {
        printf(" <option class=\"%s\">\n", cmdopt_catname[pos->cat].mnemonic );
        print_xml_help_option_element("name", pos->name1);
        print_xml_help_option_element("name", pos->name2);
        print_xml_help_option_element("name", pos->name3);
        print_xml_help_option_element("description", pos->desc);
        if (pos->eqconfig)
            print_xml_help_option_element("eqconfig", pos->eqconfig);
        else
            printf("  <eqconfig />\n");
        printf(" </option>\n");
    }
}

static void xml_help( void )
{
    printf( "<?xml version=\"1.0\"?>\n"
            "<cmdline version=\"%s\">\n", tidyReleaseDate());
    print_xml_help_option();
    printf( "</cmdline>\n" );
}

static void help( ctmbstr prog )
{
    printf( "%s [OPTIONS] input-file\n", prog );
    printf( "Utility to clean up and pretty print HTML/XHTML/XML\n");
    printf( "See http://tidy.sourceforge.net/\n");
    printf( "\n");

#ifdef PLATFORM_NAME
    printf( "Options for HTML Tidy for %s released on %s:\n",
             PLATFORM_NAME, tidyReleaseDate() );
#else
    printf( "Options for HTML Tidy released on %s:\n", tidyReleaseDate() );
#endif
    printf( "\n");

    print_help_option();

    printf( "Use \"-help-config\", or refer to the man page, for additonal configuration options.\n"
       "These options can be placed in a configuration file, in the form \"optionX: valueX\", or\n"
       "to use these configuration options on the command line, use \"--optionX valueX\"\n" );
    printf( "File output defaults to stdout, unless \"-o out-file-name\" is given.\n");
    printf( "Similarly, warnings, errors and other message default to stdout unless\n"
       "\"-f err-file\" is given.\n" );
    printf( "Single letter options, %s, may be combined, like \"-imu foo.html\".\n", single_letters);
    printf( "For further info on HTML see http://www.w3.org/MarkUp\n");
    printf( "\n");
}

static Bool isAutoBool( TidyOption topt )
{
    TidyIterator pos;
    ctmbstr def;

    if ( tidyOptGetType( topt ) != TidyInteger)
        return no;

    pos = tidyOptGetPickList( topt );
    while ( pos )
    {
        def = tidyOptGetNextPick( topt, &pos );
        if (0==strcmp(def,"yes"))
           return yes;
    }
    return no;
}

static
ctmbstr ConfigCategoryName( TidyConfigCategory id )
{
    switch( id )
    {
    case TidyMarkup:
        return "markup";
    case TidyDiagnostics:
        return "diagnostics";
    case TidyPrettyPrint:
        return "print";
    case TidyEncoding:
        return "encoding";
    case TidyMiscellaneous:
        return "misc";
    }
    fprintf(stderr, "Fatal error: impossible value for id='%d'.\n", (int)id);
    assert(0);
    abort();
    return "unknown";
}

/* Description of an option */
typedef struct {
    ctmbstr name;  /**< Name */
    ctmbstr cat;   /**< Category */
    ctmbstr type;  /**< "String, ... */
    ctmbstr vals;  /**< Potential values. If NULL, use an external function */
    ctmbstr def;   /**< default */
    tmbchar tempdefs[80]; /**< storage for default such as integer */
    Bool haveVals; /**< if yes, vals is valid */
} OptionDesc;

typedef void (*OptionFunc)( TidyDoc, TidyOption, OptionDesc * );


/* Create description "d" related to "opt" */
static
void GetOption( TidyDoc tdoc, TidyOption topt, OptionDesc *d )
{
    TidyOptionId optId = tidyOptGetId( topt );
    TidyOptionType optTyp = tidyOptGetType( topt );

    d->name = tidyOptGetName( topt );
    d->cat = ConfigCategoryName( tidyOptGetCategory( topt ) );
    d->vals = NULL;
    d->def = NULL;
    d->haveVals = yes;

    /* Handle special cases first.
     */
    switch ( optId )
    {
    case TidyDuplicateAttrs:
    case TidyNewline:
    case TidyAccessibilityCheckLevel:
        d->type = "enum";
        d->vals = NULL;
        d->def =
            optId==TidyNewline ?
            "<em>Platform dependent</em>"
            :tidyOptGetCurrPick( tdoc, optId );
        break;

    case TidyDoctype:
        d->type = "DocType";
        d->vals = NULL;
        {
            ctmbstr sdef = NULL;
            sdef = tidyOptGetCurrPick( tdoc, TidyDoctypeMode );
            if ( !sdef || *sdef == '*' )
                sdef = tidyOptGetValue( tdoc, TidyDoctype );
            d->def = sdef;
        }
        break;

    case TidyInlineTags:
    case TidyBlockTags:
    case TidyEmptyTags:
    case TidyPreTags:
        d->type = "Tag names";
        d->vals = "tagX, tagY, ...";
        d->def = NULL;
        break;

    case TidyCharEncoding:
    case TidyInCharEncoding:
    case TidyOutCharEncoding:
        d->type = "Encoding";
        d->def = tidyOptGetEncName( tdoc, optId );
        if (!d->def)
            d->def = "?";
        d->vals = NULL;
        break;

        /* General case will handle remaining */
    default:
        switch ( optTyp )
        {
        case TidyBoolean:
            d->type = "Boolean";
            d->vals = "y/n, yes/no, t/f, true/false, 1/0";
            d->def = tidyOptGetCurrPick( tdoc, optId );
            break;

        case TidyInteger:
            if (isAutoBool(topt))
            {
                d->type = "AutoBool";
                d->vals = "auto, y/n, yes/no, t/f, true/false, 1/0";
                d->def = tidyOptGetCurrPick( tdoc, optId );
            }
            else
            {
                uint idef;
                d->type = "Integer";
                if ( optId == TidyWrapLen )
                    d->vals = "0 (no wrapping), 1, 2, ...";
                else
                    d->vals = "0, 1, 2, ...";

                idef = tidyOptGetInt( tdoc, optId );
                sprintf(d->tempdefs, "%u", idef);
                d->def = d->tempdefs;
            }
            break;

        case TidyString:
            d->type = "String";
            d->vals = NULL;
            d->haveVals = no;
            d->def = tidyOptGetValue( tdoc, optId );
            break;
        }
    }
}

/* Array holding all options. Contains a trailing sentinel. */
typedef struct {
    TidyOption topt[N_TIDY_OPTIONS];
} AllOption_t;

static
int cmpOpt(const void* e1_, const void *e2_)
{
    const TidyOption* e1 = (const TidyOption*)e1_;
    const TidyOption* e2 = (const TidyOption*)e2_;
    return strcmp(tidyOptGetName(*e1), tidyOptGetName(*e2));
}

static
void getSortedOption( TidyDoc tdoc, AllOption_t *tOption )
{
    TidyIterator pos = tidyGetOptionList( tdoc );
    uint i = 0;

    while ( pos )
    {
        TidyOption topt = tidyGetNextOption( tdoc, &pos );
        tOption->topt[i] = topt;
        ++i;
    }
    tOption->topt[i] = NULL; /* sentinel */

    qsort(tOption->topt,
          /* Do not sort the sentinel: hence `-1' */
          sizeof(tOption->topt)/sizeof(tOption->topt[0])-1,
          sizeof(tOption->topt[0]),
          cmpOpt);
}

static void ForEachSortedOption( TidyDoc tdoc, OptionFunc OptionPrint )
{
    AllOption_t tOption;
    const TidyOption *topt;

    getSortedOption( tdoc, &tOption );
    for( topt = tOption.topt; *topt; ++topt)
    {
        OptionDesc d;

        GetOption( tdoc, *topt, &d );
        (*OptionPrint)( tdoc, *topt, &d );
    }
}

static void ForEachOption( TidyDoc tdoc, OptionFunc OptionPrint )
{
    TidyIterator pos = tidyGetOptionList( tdoc );

    while ( pos )
    {
        TidyOption topt = tidyGetNextOption( tdoc, &pos );
        OptionDesc d;

        GetOption( tdoc, topt, &d );
        (*OptionPrint)( tdoc, topt, &d );
    }
}

static
void PrintAllowedValuesFromPick( TidyOption topt )
{
    TidyIterator pos = tidyOptGetPickList( topt );
    Bool first = yes;
    ctmbstr def;
    while ( pos )
    {
        if (first)
            first = no;
        else
            printf(", ");
        def = tidyOptGetNextPick( topt, &pos );
        printf("%s", def);
    }
}

static
void PrintAllowedValues( TidyOption topt, const OptionDesc *d )
{
    if (d->vals)
        printf( "%s", d->vals );
    else
        PrintAllowedValuesFromPick( topt );
}

static
void printXMLDescription( TidyDoc tdoc, TidyOption topt )
{
    ctmbstr doc = tidyOptGetDoc( tdoc, topt );

    if (doc)
        printf("  <description>%s</description>\n", doc);
    else
    {
        printf("  <description />\n");
        fprintf(stderr, "Warning: option `%s' is not documented.\n",
                tidyOptGetName( topt ));
    }
}

static
void printXMLCrossRef( TidyDoc tdoc, TidyOption topt )
{
    TidyOption optLinked;
    TidyIterator pos = tidyOptGetDocLinksList(tdoc, topt);
    while( pos )
    {
        optLinked = tidyOptGetNextDocLinks(tdoc, &pos );
        printf("  <seealso>%s</seealso>\n",tidyOptGetName(optLinked));
    }
}

static
void printXMLOption( TidyDoc tdoc, TidyOption topt, OptionDesc *d )
{
    if ( tidyOptIsReadOnly(topt) )
        return;

    printf( " <option class=\"%s\">\n", d->cat );
    printf  ("  <name>%s</name>\n",d->name);
    printf  ("  <type>%s</type>\n",d->type);
    if (d->def)
        printf("  <default>%s</default>\n",d->def);
    else
        printf("  <default />\n");
    if (d->haveVals)
    {
        printf("  <example>");
        PrintAllowedValues( topt, d );
        printf("</example>\n");
    }
    else
    {
        printf("  <example />\n");
    }
    printXMLDescription( tdoc, topt );
    printXMLCrossRef( tdoc, topt );
    printf( " </option>\n" );
}

static void XMLoptionhelp( TidyDoc tdoc )
{
    printf( "<?xml version=\"1.0\"?>\n"
            "<config version=\"%s\">\n", tidyReleaseDate());
    ForEachOption( tdoc, printXMLOption );
    printf( "</config>\n" );
}

static
tmbstr GetAllowedValuesFromPick( TidyOption topt )
{
    TidyIterator pos;
    Bool first;
    ctmbstr def;
    uint len = 0;
    tmbstr val;

    pos = tidyOptGetPickList( topt );
    first = yes;
    while ( pos )
    {
        if (first)
            first = no;
        else
            len += 2;
        def = tidyOptGetNextPick( topt, &pos );
        len += strlen(def);
    }
    val = (tmbstr)malloc(len+1);
    val[0] = '\0';
    pos = tidyOptGetPickList( topt );
    first = yes;
    while ( pos )
    {
        if (first)
            first = no;
        else
            strcat(val, ", ");
        def = tidyOptGetNextPick( topt, &pos );
        strcat(val, def);
    }
    return val;
}

static
tmbstr GetAllowedValues( TidyOption topt, const OptionDesc *d )
{
    if (d->vals)
    {
        tmbstr val = (tmbstr)malloc(1+strlen(d->vals));
        strcpy(val, d->vals);
        return val;
    }
    else
        return GetAllowedValuesFromPick( topt );
}

static
void printOption( TidyDoc ARG_UNUSED(tdoc), TidyOption topt,
                  OptionDesc *d )
{
    if ( tidyOptIsReadOnly(topt) )
        return;

    if ( *d->name || *d->type )
    {
        ctmbstr pval = d->vals;
        tmbstr val = NULL;
        if (!d->haveVals)
        {
            pval = "-";
        }
        else if (pval == NULL)
        {
            val = GetAllowedValues( topt, d);
            pval = val;
        }
        print3Columns( fmt, 27, 9, 40, d->name, d->type, pval );
        if (val)
            free(val);
    }
}

static void optionhelp( TidyDoc tdoc )
{
    printf( "\nHTML Tidy Configuration Settings\n\n" );
    printf( "Within a file, use the form:\n\n" );
    printf( "wrap: 72\n" );
    printf( "indent: no\n\n" );
    printf( "When specified on the command line, use the form:\n\n" );
    printf( "--wrap 72 --indent no\n\n");

    printf( fmt, "Name", "Type", "Allowable values" );
    printf( fmt, ul, ul, ul );

    ForEachSortedOption( tdoc, printOption );
}

static
void printOptionValues( TidyDoc ARG_UNUSED(tdoc), TidyOption topt,
                        OptionDesc *d )
{
    TidyOptionId optId = tidyOptGetId( topt );
    ctmbstr ro = tidyOptIsReadOnly( topt ) ? "*" : "" ;

    switch ( optId )
    {
    case TidyInlineTags:
    case TidyBlockTags:
    case TidyEmptyTags:
    case TidyPreTags:
        {
            TidyIterator pos = tidyOptGetDeclTagList( tdoc );
            while ( pos )
            {
                d->def = tidyOptGetNextDeclTag(tdoc, optId, &pos);
                if ( pos )
                {
                    if ( *d->name )
                        printf( valfmt, d->name, d->type, ro, d->def );
                    else
                        printf( fmt, d->name, d->type, d->def );
                    d->name = "";
                    d->type = "";
                }
            }
        }
        break;
    case TidyNewline:
        d->def = tidyOptGetCurrPick( tdoc, optId );
        break;
    }

    /* fix for http://tidy.sf.net/bug/873921 */
    if ( *d->name || *d->type || (d->def && *d->def) )
    {
        if ( ! d->def )
            d->def = "";
        if ( *d->name )
            printf( valfmt, d->name, d->type, ro, d->def );
        else
            printf( fmt, d->name, d->type, d->def );
    }
}

static void optionvalues( TidyDoc tdoc )
{
    printf( "\nConfiguration File Settings:\n\n" );
    printf( fmt, "Name", "Type", "Current Value" );
    printf( fmt, ul, ul, ul );

    ForEachSortedOption( tdoc, printOptionValues );

    printf( "\n\nValues marked with an *asterisk are calculated \n"
            "internally by HTML Tidy\n\n" );
}


static void version( void )
{
   tmbstr COMP_ON = __DATE__;
   tmbstr COMP_AT = __TIME__;
#ifdef PLATFORM_NAME
    printf( "HTML Tidy for %s released on %s, compiled on %s at %s\n",
             PLATFORM_NAME, tidyReleaseDate(),
             COMP_ON, COMP_AT);
#else
    printf( "HTML Tidy released on %s, compiled on %s at %s\n", tidyReleaseDate()
       COMP_ON, COMP_AT);
#endif
}

static void unknownOption( uint c, tmbstr arg )
{
    fprintf( errout, "ERROR: UNKNOWN OPTION: %c - PART OF [%s]!\n", (char)c, arg );
}


Bool tidy_SetIndent( TidyDoc tdoc, TidyOptionId id, Bool b )
{
   Bool ret = tidyOptSetInt( tdoc, id, TidyAutoState );
   if ( tidyOptGetInt(tdoc, TidyIndentSpaces) == 0 )
      tidyOptResetToDefault( tdoc, TidyIndentSpaces );
   return ret;
}

Bool tidy_SetBool( TidyDoc tdoc, TidyOptionId id, Bool b )
{
   return tidyOptSetBool( tdoc, id, yes );
}
Bool tidy_UnsetBool( TidyDoc tdoc, TidyOptionId id, Bool b )
{
   return tidyOptSetBool( tdoc, id, no );
}

Bool tidy_Help( TidyDoc tdoc, TidyOptionId id, Bool b )
{
   help( prog );
   pgm_exit( tdoc, 0 );
   return 0; /* success */
}

Bool tidy_Xml_Help( TidyDoc tdoc, TidyOptionId id, Bool b )
{
   xml_help( );
   pgm_exit( tdoc, 0 );
   return 0; /* success */
}

Bool tidy_Conf_Help( TidyDoc tdoc, TidyOptionId id, Bool b )
{
   optionhelp( tdoc );
   pgm_exit( tdoc, 0 );
   return 0; /* success */
}

Bool tidy_Xml_Conf( TidyDoc tdoc, TidyOptionId id, Bool b )
{
   XMLoptionhelp( tdoc );
   pgm_exit( tdoc, 0 );
   return 0; /* success */
}

Bool tidy_Shw_Conf( TidyDoc tdoc, TidyOptionId id, Bool b )
{
   optionvalues( tdoc );
   pgm_exit( tdoc, 0 );
   return 0; /* success */
}

void tidy_Check_Arg2( TidyDoc tdoc, tmbstr arg2 )
{
   if ( !arg2 || ( *arg2 == 0 ) )
   {
      fprintf(errout, "ERROR: No second argment located!\n");
      pgm_exit(tdoc, 3);
   }
}

/* Set new error output stream if setting changed */
void  tidy_Fix_Error_File( TidyDoc tdoc )
{
   ctmbstr post = tidyOptGetValue( tdoc, TidyErrFile );
   if ( post && ( !errfil || !samefile(errfil, post)) )
   {
      errfil = post;
      errout = tidySetErrorFile( tdoc, post );
   }
}

Bool tidy_Conf_File( TidyDoc tdoc, tmbstr arg2 )
{
   tidy_Check_Arg2( tdoc, arg2 );
   if ( tidyLoadConfig( tdoc, arg2 ) )
   {
      fprintf(errout, "ERROR: Loading config file \"%s\"!\n", arg2);
      pgm_exit(tdoc, 3);
      return no;
   }

   /* Set new error output stream if setting changed */
   tidy_Fix_Error_File( tdoc );
   return yes;
}

Bool tidy_Set_Lang( TidyDoc tdoc, tmbstr arg2 )
{
   tidy_Check_Arg2( tdoc, arg2 );
   if ( !tidyOptSetValue( tdoc, TidyLanguage, arg2 ) )
   {
      fprintf(errout, "ERROR: Setting language \"%s\"!\n", arg2);
      pgm_exit(tdoc, 3);
      return no;
   }
   return yes;
}

Bool tidy_Set_Out( TidyDoc tdoc, tmbstr arg2 )
{
   tidy_Check_Arg2( tdoc, arg2 );
   if ( !tidyOptSetValue( tdoc, TidyOutFile, arg2 ) )
   {
      fprintf(errout, "ERROR: Setting output file \"%s\"!\n", arg2);
      pgm_exit(tdoc, 3);
      return no;
   }
   return yes;
}

Bool tidy_Set_Err( TidyDoc tdoc, tmbstr arg2 )
{
   tidy_Check_Arg2( tdoc, arg2 );
   errfil = arg2;
   errout = tidySetErrorFile( tdoc, errfil );
   return yes;
}

Bool tidy_Check_Number( tmbstr num )
{
   size_t len = strlen(num);
   size_t i;
   for ( i = 0; i < len; i++ )
   {
      int   c = num[i];
      if ( !(( c >= '0') && ( c <= '9')) )
         return no;
   }
   return yes;
}

Bool tidy_Set_Wrap( TidyDoc tdoc, tmbstr arg2 )
{
   uint wraplen = 0;
   int nfields = 0;
   tidy_Check_Arg2( tdoc, arg2 );
   if ( !tidy_Check_Number( arg2 ) )
   {
      fprintf(errout, "ERROR: Number field only required, not \"%s\"!\n", arg2);
      pgm_exit(tdoc, 3);
      return no;
   }
   nfields = sscanf( arg2, "%u", &wraplen );
   tidyOptSetInt( tdoc, TidyWrapLen, wraplen );
   return yes;
}

Bool tidy_Show_Vers( TidyDoc tdoc, tmbstr arg2 )
{
   version();
   pgm_exit( tdoc, 0 );
   return 0;  /* success */
}

Bool tidy_Set_Access( TidyDoc tdoc, tmbstr arg2 )
{
   uint acclvl = 0;
   int nfields;
   tidy_Check_Arg2( tdoc, arg2 );
   if ( !tidy_Check_Number( arg2 ) )
   {
      fprintf(errout, "ERROR: Number field only required, not \"%s\"!\n", arg2);
      pgm_exit(tdoc, 3);
      return no;
   }
   nfields = sscanf( arg2, "%u", &acclvl );
   tidyOptSetInt( tdoc, TidyAccessibilityCheckLevel, acclvl );
   return yes;
}

Bool tidy_Set_Encoding( TidyDoc tdoc, tmbstr arg )
{
   if ( tidySetCharEncoding( tdoc, arg ) ) /* == -EINVAL */
   {
      fprintf(errout, "ERROR: Setting Encosing \"%s\"!\n", arg);
      pgm_exit(tdoc, 3);
      return no;
   }
   return yes;
}

Bool tidy_NewIndent( TidyDoc tdoc, tmbstr arg2 )
{
   tidy_Check_Arg2( tdoc, arg2 );
   if ( !tidyOptParseValue(tdoc, "indent", arg2) )
   {
      fprintf(errout, "ERROR: Setting Indent \"%s\"!\n", arg2);
      pgm_exit(tdoc, 3);
      return no;
   }
   return yes;
}

#define  TIDY_MAX_ARGS  128


int   Process_Args( TidyDoc tdoc, int argc, char** argv );
int   Process_Input( TidyDoc tdoc, char * file )
{
   FILE * fp;
   struct stat sbuf;
   tmbstr fbuf;
   ulong size, ui, len, bgnui;
   tmbstr * argv;
   int   argc;
   tmbstr arg, p;
   char c;
   if ( stat( file, &sbuf ) )
   {
      fprintf(errout, "ERROR: stat of input file \"%s\" failed!\n", file);
      pgm_exit(tdoc, 3);
   }
   if ( sbuf.st_size == 0 )
      return 0;   /* quietly ignore zero length files */
   if ( sbuf.st_size > 4294967295 )
   {
      fprintf(errout, "ERROR: input file \"%s\" TOO LARGE!\n", file);
      pgm_exit(tdoc, 3);
   }
   fbuf = malloc( sbuf.st_size + (sizeof(char *) * (TIDY_MAX_ARGS + 1)) );
   if ( !fbuf )
   {
      fprintf(errout, "ERROR: memory allocation for input file \"%s\" failed!\n", file);
      pgm_exit(tdoc, 3);
   }
   fp = fopen( file, "rb" );
   if ( !fp )
   {
      free( fbuf );
      fprintf(errout, "ERROR: Opening input file \"%s\" failed!\n", file);
      pgm_exit(tdoc, 3);
   }
   size = fread( fbuf, 1, sbuf.st_size, fp );
   fclose( fp );
   if ( size != sbuf.st_size )
   {
      free( fbuf );
      fprintf(errout, "ERROR: Full read of input file \"%s\" failed!\n", file);
      pgm_exit(tdoc, 3);
   }
   argv = (char **)&fbuf[size];
   argc = 0;
   argv[argc++] = prog;
   for ( ui = 0; ui < size; ui++ )
   {
      c = fbuf[ui];
      if ( c > ' ' )
      {
         if ( c == ';' )
         {
            ui++;
            for ( ; ui < size; ui++ )
            {
               c = fbuf[ui];
               if (( c < ' ' ) && ( c != '\t' ))
                  break;
            }
         }
         else
         {
            arg = &fbuf[ui];
            bgnui = ui;
            ui++;
            len = 1;
            for ( ; ui < size; ui++ )
            {
               c = fbuf[ui];
               if ((( c < ' ' ) && ( c != '\t' )) || ( c == ';' )) {
                  fbuf[ui] = 0;  /* zero termination */
                  if ( c == ';' )
                  {
                     ui++;
                     for ( ; ui < size; ui++ )
                     {
                        c = fbuf[ui];
                        if (( c < ' ' ) && ( c != '\t' ))
                           break;
                     }
                  }
                  break;
               }
               len++;
            }
            while(len--)
            {
               p = &fbuf[bgnui+len];
               if ( *p > ' ' )
                  break;
               *p = 0;
            }
            if ( *arg == '"' )
            {
               /* unpeel the quotes */
               arg++;
               p = strrchr(arg, '"');
               if(p)
                  *p = 0;
            }
            else if ( *arg == '-' )
            {
               /* may be two! */
               p = strchr(arg, ' ');
               if ( p )
               {
                  *p = 0;
                  argv[argc++] = arg;
                  p++;  /* step up to next char */
                  while ( *p <= ' ' )
                     p++;  /* walk of spacey stuff */
                  arg = p;
               }
            }
            argv[argc++] = arg;
            if ( argc >= (TIDY_MAX_ARGS - 1) )
            {
               free( fbuf );
               fprintf(errout, "ERROR: TOO MANY ARGUMENTS IN file \"%s\" failed!\n"
                  "RECOMPILE increasing TIDY_MAX_ARGS!\n", file);
               pgm_exit(tdoc, 3);
            }
         }
      }
   }
   argv[argc] = 0;   /* NULL terminate array of pointers */
   Process_Args( tdoc, argc, argv );
   free( fbuf );
   return 0;
}

static tmbstr tidy_strdupe( TidyDoc tdoc, tmbstr stg )
{
   ulong len = strlen(stg);
   tmbstr dupe = (tmbstr)malloc( len + 1 );
   if ( !dupe )
   {
      fprintf(errout, "ERROR: MEMORY FAILED!\n" );
      pgm_exit( tdoc, 3 );
   }
   strcpy(dupe, stg);
   return dupe;
}

tmbstr tidy_Get_Name3( const CmdOptDesc * COD )
{
   static tmbchar cname3[64];
   tmbstr name = cname3;
   tmbstr p;
   if ( COD->name3 )
   {
      strcpy(name, COD->name3);
      p = strchr(name, ' ');
      if ( p )
         *p = 0;
      return name;
   }
   return NULL;
}

tmbstr tidy_Get_Name2( const CmdOptDesc * COD )
{
   static tmbchar cname2[64];
   tmbstr name = cname2;
   tmbstr p;
   if ( COD->name2 )
   {
      strcpy(name, COD->name2);
      p = strchr(name, ' ');
      if ( p )
         *p = 0;
      return name;
   }
   return NULL;
}


tmbstr tidy_Get_Name1( const CmdOptDesc * COD )
{
   static tmbchar cname1[64];
   tmbstr name = cname1;
   tmbstr p;
   strcpy(name, COD->name1);
   p = strchr(name, ' ');
   if ( p )
      *p = 0;
   return name;
}

CmdOptDesc * Tidy_Get_Cmd_Opt_Ptr( tmbstr arg )
{
   tmbstr name1, name2, name3;
   CmdOptDesc * COD = &cmdopt_defs[0];
   while( COD->name1 )
   {
      name1 = tidy_Get_Name1( COD );
      name2 = tidy_Get_Name2( COD );
      name3 = tidy_Get_Name3( COD );
      if (( name1 && ( strcasecmp(arg, name1) == 0 )) ||
          ( name2 && ( strcasecmp(arg, name2) == 0 )) ||
          ( name3 && ( strcasecmp(arg, name3) == 0 )) )
      {
         return COD;
      }
      COD++;
   }
   return NULL;
}

int   Process_Args( TidyDoc tdoc, int argc, char** argv )
{
   static tmbstr  sargv[3];
   int   i, dncmd;
   Bool is2;
   tmbstr arg, arg2;
   tmbchar sargs[3];
   tmbstr  sarg = sargs;
   const CmdOptDesc * COD;
#ifndef  NDEBUG
   tmbstr larg, larg2;
   arg = arg2 = larg = larg2 = 0;
#endif /* #ifndef  NDEBUG */
   dncmd = 0;
/* #ifndef  NDEBUG */
   save_commands( tdoc, argc, argv );
/* #endif /* #ifndef  NDEBUG */

   for ( i = 1; i < argc; i++ ) {
#ifndef  NDEBUG
      larg = arg;
      larg2 = arg2;
#endif /* #ifndef  NDEBUG */
      dncmd = 0;
      is2  = no;
      arg = argv[i];
      arg2 = 0;
      if( (i + 1) < argc )
         arg2 = argv[i+1];

      /* [ tidy-Feature Requests-1867576 ]
         Windows command line parameters are not treated correctly */
      if ( *arg == '/' ) { /* if windows '/' argument */
         unknownOption( *arg, arg );
         tidy_Help( tdoc, 0, 0 );
      }

      if ( *arg == '@' )
      {
         arg++;
         Process_Input( tdoc, arg );
      }
      else if( *arg == '-' )
      {
         /* fully support -foo or --foo for all commands */
         if ( arg[1] == '-' )
         {
#ifdef  TIDY_NEW_INDENT
            arg++;
            is2 = yes;
#else /* !#ifdef  TIDY_NEW_INDENT */
            if ( strcasecmp(arg, "--indent") )
            {
               arg++;
               is2 = yes;
            }
#endif /* #ifdef  TIDY_NEW_INDENT y/n */
         }
         if (( COD = Tidy_Get_Cmd_Opt_Ptr( arg ) ) != NULL )
         {
            if( COD->cmdtable.servb )
            {
               COD->cmdtable.servb( tdoc, COD->cmdtable.optid, is2 );
            }
            else if ( COD->cmdtable.serv2 )
            {
               COD->cmdtable.serv2( tdoc, arg2 );
               i++;  /* used up argument 2 */
            }
            else if ( COD->cmdtable.serv3 )
            {
               arg++;
               COD->cmdtable.serv3( tdoc, arg );
            }
            dncmd = 1;
         }

         if ( !dncmd )
         {
            arg++;
            if ( *arg == '-' )
            {
               is2 = yes;
               arg++;
            }
            if ( is2 )
            {
               /* else if ( strncmp(argv[1], "--", 2 ) == 0) */
               if ( tidyOptParseValue(tdoc, arg, arg2) )
               {
                  /* Set new error output stream if setting changed */
                  tidy_Fix_Error_File( tdoc );
                  dncmd = 1;
                  i++; /* used up argument */
               }
               /* else fall into !dncmd below == ERROR */
            }
            else if ( argv != &sargv[0] )
            {
               /* can only try for list of single options
                  cases
                  'i' - TidyIndentContent to auto(and reset TidyIndentSpaces),
                  'u' - TidyUpperCaseTags,
                  'c' - TidyMakeClean,
                  'b' - TidyMakeBare,
                  'n' - TidyNumEntities,
                  'm' - TidyWriteBack,
                  'e' - TidyShowMarkup (no),
                  'q' - TidyQuiet
               */
               uint c;
               ctmbstr s = arg;
               sarg[0] = '-';
               sarg[2] = 0;
               sargv[0] = prog;
               sargv[2] = 0;
               while ( c = *s++ )
               {
                  if ( !strchr( single_letters, c ) )
                  {
                     unknownOption( c, argv[i] );
                     pgm_exit( tdoc, 3 );
                  }
               }
               s = arg;
               while ( c = *s++ )
               {
                  sarg[1] = (tmbchar)c;
                  sargv[1] = sarg;
                  Process_Args( tdoc, 2, &sargv[0] );
               }
               dncmd = 1;
            }
         }

         if ( !dncmd )
         {
            fprintf( errout, "ERROR: Unknown command: %s\n", argv[i] );
            pgm_exit( tdoc, 3 );
         }
      }
      else
      {
         /* not '@' or '-', then assumed is the HTML input file */
         if ( htmlfil )
         {
            fprintf( errout, "ERROR: Found second input file: %s\n"
               "Already have %s!\n", arg, htmlfil );
            pgm_exit( tdoc, 3 );
         }

         htmlfil = tidy_strdupe(tdoc, arg); /* duplicate string */
         /* printf( "Set out file to %s ... (%s)\n", htmlfil, arg ); */
      }
   }
   return 0;
}

/* *****************************************************
 * void tidyRenameFile( TidyDoc tdoc, ctmbstr htmlfil )
 *
 * Just can NOT be too CAUTIOUS!!!
 * We are about to OVERWRITE the existing file,
 * so I want to RENAME it FIRST
 * I use my 'time honoured' method, of -
 * If no .old, rename to .old,
 * Else rename to .bak, deleting any previous .bak
 * ***************************************************** */
void tidyRenameFile( TidyDoc tdoc, ctmbstr htmlfil )
{
   static tmbchar bak[264];
   struct stat sbuf;
   tmbstr bb = bak;
   size_t len;
   tmbstr  p;

   strcpy(bb, htmlfil);
   len = strlen(bb);
   p = strrchr(bb,'.'); /* get LAST '.' */
   if ( !p )   /* if none, */
      p = &bb[len];  /* append */
   strcpy(p, ".old");   /* first choice - rename to .old */
   if ( stat( bb, &sbuf ) == 0 )
   {
      /* .old already EXISTS, change to .bak */
      strcpy(p, ".bak");
      if ( stat( bb, &sbuf ) == 0 )
      {
         /* if .bak exist, DELETE it */
         remove(bb);
      }
   }
   rename( htmlfil, bb );

}

void tidy_Show_CWD( void )
{
   static tmbchar cwd[264];
   _getcwd( cwd, sizeof(cwd) );
   fprintf(errout, "Current Work Directory: [%s]\n", cwd );
}

int main( int iargc, char** cargv )
{
    /* ctmbstr cfgfil = NULL, errfil = NULL, htmlfil = NULL; */
    TidyDoc tdoc = tidyCreate();
    int status = 0;

    uint contentErrors = 0;
    uint contentWarnings = 0;
    uint accessWarnings = 0;

    errout = stderr;  /* initialize to stderr */
    status = 0;
    prog = cargv[0];

#ifdef TIDY_CONFIG_FILE
    if ( tidyFileExists( tdoc, TIDY_CONFIG_FILE) )
    {
        status = tidyLoadConfig( tdoc, TIDY_CONFIG_FILE );
        if ( status != 0 ) {
           fprintf(errout, "ERROR: Loading config file \"%s\" failed, err = %d\n", TIDY_CONFIG_FILE, status);
           pgm_exit( tdoc, 3 );
           return 3;
        }
    }
#endif /* TIDY_CONFIG_FILE */

    /* look for env var "HTML_TIDY" */
    /* then for ~/.tidyrc (on platforms defining $HOME) */
#if   WIN32
    if ( ( homedrive = getenv("HOMEDRIVE") ) &&
         ( homepath = getenv("HOMEPATH") ) )
    {
       strcpy(tidyrc, homedrive);
       strcat(tidyrc, homepath);
       strcat(tidyrc, "\\.tidyrc");
       status = tidyLoadConfig( tdoc, tidyrc );
    }
#endif   /* WIN32 */


    if ( cfgfil = getenv("HTML_TIDY") )
    {
        status = tidyLoadConfig( tdoc, cfgfil );
        if ( status != 0 ) {
           fprintf(errout, "ERROR: Loading config file \"%s\" failed, err = %d\n", cfgfil, status);
           pgm_exit( tdoc, 3 );
           return 3;
        }
    }
#ifdef TIDY_USER_CONFIG_FILE
    else if ( tidyFileExists( tdoc, TIDY_USER_CONFIG_FILE) )
    {
        status = tidyLoadConfig( tdoc, TIDY_USER_CONFIG_FILE );
        if ( status != 0 ) {
           fprintf(errout, "ERROR: Loading config file \"%s\" failed, err = %d\n", TIDY_USER_CONFIG_FILE, status);
           pgm_exit( tdoc, 3 );
           return 3;
        }
    }
#endif /* TIDY_USER_CONFIG_FILE */

    /* read command line */
    Process_Args( tdoc, iargc, cargv );

#ifdef   WIN32
    if ( !htmlfil )
    {
       fprintf(errout, "ERROR: No input file located in command!\n" );
       pgm_exit( tdoc, 3 );
       return 3;
    }

    tidyOptSetValue( tdoc, TidyEmacsFile, htmlfil ); 
    /* printf( "Parsing file %s ...\n", htmlfil ); */
    status = tidyParseFile( tdoc, htmlfil );
    if ( status < 0 )
    {
       if ( status == -ENOENT )
       {
          tidy_Show_CWD();
          fprintf(errout, "ERROR: FILE %s NOT FOUND! (%d)\n", htmlfil, status );
       }
       else if ( status == -ENOMEM )
          fprintf(errout, "ERROR: FILE %s MAPPING FAILED! (%d)\n", htmlfil, status );
       else
          fprintf(errout, "ERROR: FILE %s PARSE FAILED! (%d)\n", htmlfil, status );
       pgm_exit( tdoc, 3 );
       return 3;
    }
#else /* !#ifdef   WIN32 */
   if ( htmlfil )
   {
      if ( tidyOptGetBool(tdoc, TidyEmacs) )
         tidyOptSetValue( tdoc, TidyEmacsFile, htmlfil );
      status = tidyParseFile( tdoc, htmlfil );
   }
   else
   {
      htmlfil = strdup("stdin");
      status = tidyParseStdin( tdoc );
   }
#endif /* #ifdef   WIN32 y/n */

   if ( status >= 0 ) {
#ifndef  NDEBUG
      printf( "Doing tidyCleanAndRepair ...\n" );
#endif   /* #ifndef  NDEBUG */
       status = tidyCleanAndRepair( tdoc );
   }
   if ( status >= 0 ) {
#ifndef  NDEBUG
      printf( "Doing tidyRunDiagnostics ...\n" );
#endif   /* #ifndef  NDEBUG */
       status = tidyRunDiagnostics( tdoc );
   }

    if ( status > 1 ) /* If errors, do we want to force output? */
       status = ( tidyOptGetBool(tdoc, TidyForceOutput) ? status : -1 );

    if ( status >= 0 && tidyOptGetBool(tdoc, TidyShowMarkup) )
    {
       if ( tidyOptGetBool(tdoc, TidyWriteBack) && htmlfil )
       {
#ifndef  NDEBUG
          printf( "Doing tidyRenameFile ...\n" );
#endif   /* #ifndef  NDEBUG */
          tidyRenameFile( tdoc, htmlfil );
#ifndef  NDEBUG
          printf( "Doing tidySaveFile to %s ...\n", htmlfil );
#endif   /* #ifndef  NDEBUG */
          status = tidySaveFile( tdoc, htmlfil );
       }
       else
       {
          ctmbstr outfil = tidyOptGetValue( tdoc, TidyOutFile );
          if ( outfil )
          {
#ifndef  NDEBUG
             printf( "Doing tidySaveFile to %s ...\n", outfil );
#endif   /* #ifndef  NDEBUG */
             status = tidySaveFile( tdoc, outfil );
             if ( status == -ENOENT )
             {
                tidy_Show_CWD();
                fprintf(errout, "ERROR: FILE %s COULD NOT BE CREATED! (%d)\n", outfil, status );
                pgm_exit( tdoc, 3 );
                return 3;
             }
          }
          else
          {
#ifndef  NDEBUG
             printf( "Doing tidySaveStdout ...\n" );
#endif   /* #ifndef  NDEBUG */
             status = tidySaveStdout( tdoc );
          }
       }
    }

    contentErrors   += tidyErrorCount( tdoc );
    contentWarnings += tidyWarningCount( tdoc );
    accessWarnings  += tidyAccessWarningCount( tdoc );

    if (!tidyOptGetBool(tdoc, TidyQuiet) &&
        errout == stderr && !contentErrors)
        fprintf(errout, "\n");

    if (contentErrors + contentWarnings > 0 && 
         !tidyOptGetBool(tdoc, TidyQuiet))
        tidyErrorSummary(tdoc);

 /* turn off this REPEATED information ...
    if (!tidyOptGetBool(tdoc, TidyQuiet))
        tidyGeneralInfo(tdoc);
    ======================================= */

    /* return status can be used by scripts */
    if ( contentErrors > 0 )
        pgm_exit( tdoc, 2 );

    if ( contentWarnings > 0 )
        pgm_exit( tdoc, 1 );

    /* 0 signifies all is ok */
    pgm_exit( tdoc, 0 );
    return 0;
}

// eof - tidy2.c
